//.          GPS module for "Virtual Town"project
//.                                                       
//.  Authors: PAV <AlxPonom@mail.ru>             
//.                                                       
//.  This program is free software under the GPL (>= v2)  
//.  Read the file COPYING coming with project for details 

using System;

namespace CrdSys
{
	using GEOGraphServer.TTGeodesyPointSOAPFunctionality;

	/// <summary>
	/// Crd-Sys converter
	/// </summary>
	public class TCrdSysConvertor
	{
		private string ServerURL;
		private string UserName;
		private string UserPassword;
		private int idCrdSys;
		private int PointsCount;
		private byte[] GeoToXY_Points = null;
		public double GeoToXY_Points_Lat;
		public double GeoToXY_Points_Long;

		public TCrdSysConvertor(string pServerURL, string pUserName, string pUserPassword, int pidCrdSys, int pPointsCount)
		{
			ServerURL = pServerURL;
			UserName = pUserName;
			UserPassword = pUserPassword;
			idCrdSys = pidCrdSys;
			PointsCount = pPointsCount;
		}

		public bool GeoToXY_LoadPoints(double pLat, double pLong)
		{
			ITTGeodesyPointSOAPFunctionalityService GeodesyPointType = new ITTGeodesyPointSOAPFunctionalityService(ServerURL);
			//.
			if (!GeodesyPointType.CrdSys_GetNearestGeoPoints(UserName,UserPassword,idCrdSys, pLat,pLong, PointsCount, out GeoToXY_Points)) return false; //. ->
			GeoToXY_Points_Lat = pLat;
			GeoToXY_Points_Long = pLong;
			return true;
		}

		public bool ConvertGeoToXY(double Lat, double Long, out double Xr, out double Yr)
		{
			if (GeoToXY_Points != null)
			{
				double SumX = 0;
				double SumY = 0;
				int Count = 0;
				TGeodesyPointStruct P0 = new TGeodesyPointStruct(), P1 = new TGeodesyPointStruct();
				int PointIndex = 0;
				for (int I = 0; I < PointsCount; I++)
				{
					P0.PrepareFromBytes(GeoToXY_Points, ref PointIndex);
					int PairPointIndex = PointIndex;
					for (int J = (I+1); J < PointsCount; J++)
					{
						P1.PrepareFromBytes(GeoToXY_Points, ref PairPointIndex);
						double X,Y;
						_ConvertGeoToXY(P0,P1, Lat,Long,  out X,out Y);
						SumX = SumX+X; SumY = SumY+Y;
						Count++;
					};
				};
				Xr = (SumX/Count); Yr = (SumY/Count);
				return true; //. ->
			}
			else
			{
				Xr = 0;
				Yr = 0;
				return false;
			};
		}

		public double GetDistance(double StartLat, double StartLong,  double EndLat, double EndLong)
		{
			double fPhimean;
			double fdLambda;
			double fdPhi;
			double fAlpha;
			double fRho;
			double fNu;
			double fR;
			double fz;
			double fTemp;
			double Distance;
			double Bearing;
			//. consts
			double D2R = 0.017453;
			double R2D = 57.295781;
			double a = 6378137.0; 
			double e2 = 0.006739496742337; 

			fdLambda = (StartLong - EndLong) * D2R;
			fdPhi = (StartLat - EndLat) * D2R;
			fPhimean = ((StartLat + EndLat) / 2.0) * D2R;

			fTemp = 1 - e2 * (Math.Pow(Math.Sin(fPhimean), 2));
			fRho = (a * (1 - e2)) / Math.Pow(fTemp, 1.5);
			fNu = a / (Math.Sqrt(1 - e2 * (Math.Sin(fPhimean) * Math.Sin(fPhimean))));
			fz = Math.Sqrt(Math.Pow(Math.Sin(fdPhi / 2.0), 2) + Math.Cos(EndLat * D2R) * Math.Cos(StartLat * D2R) *	Math.Pow(Math.Sin(fdLambda / 2.0), 2));
			fz = 2 * Math.Asin(fz);

			fAlpha = Math.Cos(EndLat * D2R) * Math.Sin(fdLambda) * 1 / Math.Sin(fz);
			fAlpha = Math.Asin(fAlpha);
			
			fR = (fRho * fNu) / ((fRho * Math.Pow(Math.Sin(fAlpha), 2)) + (fNu *	Math.Pow(Math.Cos(fAlpha), 2)));
			Distance = (fz * fR);

			if ((StartLat < EndLat) && (StartLong < EndLong)) 
			{
				Bearing = Math.Abs(fAlpha * R2D);
			}
			else if ((StartLat < EndLat) && (StartLong > EndLong)) 
			{
				Bearing = 360 - Math.Abs(fAlpha * R2D);
			}
			else if ((StartLat > EndLat) && (StartLong > EndLong)) 
			{
				Bearing = 180 + Math.Abs(fAlpha * R2D);
			}
			else if ((StartLat > EndLat) && (StartLong < EndLong)) 
			{
				Bearing = 180 - Math.Abs(fAlpha * R2D);
			};
			return Distance;
		}
		private void _ConvertGeoToXY(TGeodesyPointStruct P0, TGeodesyPointStruct P1, double Lat, double Long, out double Xr, out double Yr)
		{
			double LatD,LongD;
			double Aspect;
			double dX,dY;
			double dLat,dLong;
			double L,LL;
			double Scale;
			double Alpha,Betta;
			double Angle;
			double TranslateX,TranslateY;
			double cfTransMeter = 10000.0;
		
			//. get sector aspect
			LatD = GetDistance(Lat,Long,Lat+1/60.0,Long);
			LongD = GetDistance(Lat,Long,Lat,Long+1/60.0);
			Aspect = (LongD/LatD);
			//. do linear geo coordinates
			P0.Longitude = P0.Longitude*Aspect;
			P1.Longitude = P1.Longitude*Aspect;
			Long = Long*Aspect;
			//.
			dX = (P1.X-P0.X)*cfTransMeter; dY = (P1.Y-P0.Y)*cfTransMeter;
			dLong = (P1.Longitude-P0.Longitude); dLat = (P1.Latitude-P0.Latitude);
			//. transform scale
			L = Math.Pow(dX,2)+Math.Pow(dY,2);
			LL = Math.Pow(dLong,2)+Math.Pow(dLat,2);
			Scale = Math.Sqrt(L/LL);
			//. transform angle
			if (dX != 0)
			{
				Alpha = Math.Atan(dY/dX);
				if (dX < 0) Alpha = Alpha+Math.PI;
			}
			else
			{
				if (dY >= 0) Alpha = Math.PI/2; else Alpha = -Math.PI/2;
			};
			if (dLong != 0)
			{
				Betta = Math.Atan(dLat/dLong);
				if (dLong < 0)Betta = Betta+Math.PI;
			}
			else
			{
				if (dLat >= 0)Betta = Math.PI/2; else Betta = -Math.PI/2;
			}
			Angle = (Alpha-Betta);
			//. transform translate
			TranslateX = P0.X*cfTransMeter; TranslateY = P0.Y*cfTransMeter;
			RotateByAngleFrom0(ref TranslateX, ref TranslateY, (-Angle));
			TranslateX = (TranslateX-P0.Longitude*Scale); TranslateY = (TranslateY-P0.Latitude*Scale);
			//. transforming from (Lat;Long) to (X;Y)
			Xr = TranslateX+Long*Scale; Yr = TranslateY+Lat*Scale;
			RotateByAngleFrom0(ref Xr, ref Yr, Angle);
			Xr = Xr/cfTransMeter; Yr = Yr/cfTransMeter;
		}

		private void RotateByAngleFrom0(ref double X, ref double Y, double Angle)
		{	
			double _X,_Y;
			_X = X*Math.Cos(Angle)+Y*(-Math.Sin(Angle));
			_Y = X*Math.Sin(Angle)+Y*Math.Cos(Angle);
			X = _X; Y = _Y;
		}

	}

	public struct TGeodesyPointStruct
	{
		public double X;
		public double Y;
		public double Latitude;
		public double Longitude;

		public void PrepareFromBytes(byte[] Bytes, ref int Idx)
		{
			X = BitConverter.ToDouble(Bytes,Idx); Idx+=8;
			Y = BitConverter.ToDouble(Bytes,Idx); Idx+=8;
			Latitude = BitConverter.ToDouble(Bytes,Idx); Idx+=8;
			Longitude = BitConverter.ToDouble(Bytes,Idx); Idx+=8;
		}
	}
}
