//.          GPS module for "Virtual Town"project
//.                  Copyright (c) 2005 
//.                                                       
//.  Authors: PAV <AlxPonom@mail.ru>             
//.                                                       
//.  This program is free software under the GPL (>= v2)  
//.  Read the file COPYING coming with project for details 

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Security.Cryptography;
using System.Text;
using System.Runtime.InteropServices;

namespace GEOGraphServer
{
	using GEOGraphServer.TTGeodesyPointSOAPFunctionality;
	using GEOGraphServer.TBase2DVisualizationSOAPFunctionality;
	using CrdSys;

	delegate void MeasurementReceivedEvent(GEOGraphMessage Message);

	/// <summary>
	/// Summary description for DevicesServer.
	/// </summary>
	public class TfmDevicesServer : System.Windows.Forms.Form
	{
		private static int ProcessPointsCount = 3;
		public static int MaxDistanceToReloadProcessingPoints = 1000; //. meters
		private DevicesServer devicesServer = null;
		private System.Windows.Forms.ListView lvLog;
		private System.Windows.Forms.ColumnHeader columnHeader1;
		private System.Windows.Forms.ColumnHeader columnHeader2;
		private System.Windows.Forms.ColumnHeader columnHeader3;
		private System.Windows.Forms.ColumnHeader columnHeader4;
		private System.Windows.Forms.Button btnShowUsers;
		private System.Windows.Forms.CheckBox cbLogEnabled;
		private System.Windows.Forms.CheckBox cbProcessing;
		private System.Windows.Forms.ColumnHeader colunmHeader5;

		/// <summary>
		/// Required designer variable.
		/// </summary>
		private System.ComponentModel.Container components = null;

		public TfmDevicesServer()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();
			//. init server
			devicesServer = new DevicesServer(this);
			//.
			Text = "GPS devices server (Port: "+devicesServer.Configuration.Port.ToString()+", MaxConnections: "+devicesServer.Configuration.MaxConnections.ToString()+")";
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			if (devicesServer != null) 
			{
					devicesServer.Dispose();
					devicesServer = null;
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.lvLog = new System.Windows.Forms.ListView();
			this.columnHeader1 = new System.Windows.Forms.ColumnHeader();
			this.columnHeader2 = new System.Windows.Forms.ColumnHeader();
			this.columnHeader3 = new System.Windows.Forms.ColumnHeader();
			this.columnHeader4 = new System.Windows.Forms.ColumnHeader();
			this.btnShowUsers = new System.Windows.Forms.Button();
			this.cbLogEnabled = new System.Windows.Forms.CheckBox();
			this.cbProcessing = new System.Windows.Forms.CheckBox();
			this.colunmHeader5 = new System.Windows.Forms.ColumnHeader();
			this.SuspendLayout();
			// 
			// lvLog
			// 
			this.lvLog.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
			this.lvLog.Columns.AddRange(new System.Windows.Forms.ColumnHeader[] {
																					this.columnHeader1,
																					this.colunmHeader5,
																					this.columnHeader2,
																					this.columnHeader3,
																					this.columnHeader4});
			this.lvLog.Dock = System.Windows.Forms.DockStyle.Top;
			this.lvLog.Enabled = false;
			this.lvLog.Font = new System.Drawing.Font("Tahoma", 10F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(204)));
			this.lvLog.FullRowSelect = true;
			this.lvLog.GridLines = true;
			this.lvLog.HideSelection = false;
			this.lvLog.Location = new System.Drawing.Point(0, 0);
			this.lvLog.MultiSelect = false;
			this.lvLog.Name = "lvLog";
			this.lvLog.Size = new System.Drawing.Size(878, 128);
			this.lvLog.Sorting = System.Windows.Forms.SortOrder.Ascending;
			this.lvLog.TabIndex = 3;
			this.lvLog.View = System.Windows.Forms.View.Details;
			// 
			// columnHeader1
			// 
			this.columnHeader1.Text = "Time";
			this.columnHeader1.Width = 100;
			// 
			// columnHeader2
			// 
			this.columnHeader2.Text = "User";
			this.columnHeader2.Width = 100;
			// 
			// columnHeader3
			// 
			this.columnHeader3.Text = "Position";
			this.columnHeader3.Width = 300;
			// 
			// columnHeader4
			// 
			this.columnHeader4.Text = "Precision";
			this.columnHeader4.Width = 150;
			// 
			// btnShowUsers
			// 
			this.btnShowUsers.Font = new System.Drawing.Font("Microsoft Sans Serif", 10F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(204)));
			this.btnShowUsers.Location = new System.Drawing.Point(192, 137);
			this.btnShowUsers.Name = "btnShowUsers";
			this.btnShowUsers.Size = new System.Drawing.Size(120, 32);
			this.btnShowUsers.TabIndex = 4;
			this.btnShowUsers.Text = "Show Users";
			this.btnShowUsers.Click += new System.EventHandler(this.btnShowUsers_Click);
			// 
			// cbLogEnabled
			// 
			this.cbLogEnabled.Font = new System.Drawing.Font("Microsoft Sans Serif", 10F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(204)));
			this.cbLogEnabled.Location = new System.Drawing.Point(120, 140);
			this.cbLogEnabled.Name = "cbLogEnabled";
			this.cbLogEnabled.Size = new System.Drawing.Size(64, 24);
			this.cbLogEnabled.TabIndex = 5;
			this.cbLogEnabled.Text = "Log";
			this.cbLogEnabled.CheckedChanged += new System.EventHandler(this.cbLogEnabled_CheckedChanged);
			// 
			// cbProcessing
			// 
			this.cbProcessing.Checked = true;
			this.cbProcessing.CheckState = System.Windows.Forms.CheckState.Checked;
			this.cbProcessing.Font = new System.Drawing.Font("Microsoft Sans Serif", 10F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(204)));
			this.cbProcessing.Location = new System.Drawing.Point(8, 140);
			this.cbProcessing.Name = "cbProcessing";
			this.cbProcessing.Size = new System.Drawing.Size(96, 24);
			this.cbProcessing.TabIndex = 6;
			this.cbProcessing.Text = "Processing";
			this.cbProcessing.CheckedChanged += new System.EventHandler(this.cbProcessing_CheckedChanged);
			// 
			// colunmHeader5
			// 
			this.colunmHeader5.Text = "Object";
			this.colunmHeader5.Width = 200;
			// 
			// TfmDevicesServer
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(878, 177);
			this.Controls.Add(this.cbProcessing);
			this.Controls.Add(this.cbLogEnabled);
			this.Controls.Add(this.btnShowUsers);
			this.Controls.Add(this.lvLog);
			this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.Fixed3D;
			this.MaximizeBox = false;
			this.Name = "TfmDevicesServer";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
			this.ResumeLayout(false);

		}
		#endregion

		private void UpdateOnMeasurementReceived(GEOGraphMessage Message)
		{
			if (!cbLogEnabled.Checked) return; //. ->
			TGEOGraphUser User;
			TGEOGraphUserObject Object;
			string UserName = null;
			string ObjectName = null;
			lock(devicesServer.Users)
			{
				if (devicesServer.Users.GetUserAndObjectByObjectID(Message.ObjectID, out User, out Object)) 
				{
					UserName = User.UserName;
					ObjectName = Object.ObjectName;
				}
			};
			//.
			if (UserName != null)
			{
				ListViewItem item = new ListViewItem(DateTime.Now.ToShortTimeString(),1);
				item.SubItems.Add(ObjectName);
				item.SubItems.Add(UserName);
				item.SubItems.Add(Message.Latitude.ToString()+"; "+Message.Longitude.ToString());
				item.SubItems.Add(Message.Precision.ToString());
				lvLog.Items.Add(item);
				item.EnsureVisible();
				item.Focused = true;
				item.Selected = true;
			};
		}

		public void DoOnMeasurementReceived(GEOGraphMessage Message)
		{
			bool flProcessing = cbProcessing.Checked;
			lock(devicesServer.Users)
			{
				TGEOGraphUser User;
				TGEOGraphUserObject Object;
				if (devicesServer.Users.GetUserAndObjectByObjectID(Message.ObjectID, out User, out Object) && flProcessing)
				{
					//. move visualization
					if (Object.CrdSysConvertor == null)
					{
						Object.CrdSysConvertor = new TCrdSysConvertor(devicesServer.Configuration.SpaceSOAPServerURL,User.UserName,User.UserPassword,Object.GeoCrdSysID,ProcessPointsCount);
						Object.CrdSysConvertor.GeoToXY_LoadPoints(Message.Latitude,Message.Longitude);
						//. get object base points
						ITBase2DVisualizationSOAPFunctionalityService V = new ITBase2DVisualizationSOAPFunctionalityService(devicesServer.Configuration.SpaceSOAPServerURL);
						byte[] BA = V.GetNodes(User.UserName,User.UserPassword,Object.idTVisualization,Object.idVisualization);
						int Idx = 0;
						int Count = BitConverter.ToInt32(BA, Idx); Idx+=4;
						if (Count == 2) 
						{
							Object.X1 = BitConverter.ToDouble(BA, Idx); Idx+=8;
							Object.Y1 = BitConverter.ToDouble(BA, Idx); Idx+=8;
							Object.X0 = BitConverter.ToDouble(BA, Idx); Idx+=8;
							Object.Y0 = BitConverter.ToDouble(BA, Idx); Idx+=8;
							Object.L = Math.Pow((Object.X1-Object.X0),2)+Math.Pow((Object.Y1-Object.Y0),2);
						}
						else
						{
							Object.X1 = 0;
							Object.Y1 = 0;
							Object.X0 = 0;
							Object.Y0 = 0;
							Object.L = 0;
						};
					}
					else
					{
						double Distance = Object.CrdSysConvertor.GetDistance(Object.CrdSysConvertor.GeoToXY_Points_Lat,Object.CrdSysConvertor.GeoToXY_Points_Long,Message.Latitude,Message.Longitude);
						if (Distance > MaxDistanceToReloadProcessingPoints)
						{
							Object.CrdSysConvertor.GeoToXY_LoadPoints(Message.Latitude,Message.Longitude);
						};
					};
					double X,Y;
					Object.CrdSysConvertor.ConvertGeoToXY(Message.Latitude,Message.Longitude, out X, out Y);
					//. move & rotate visualization
					double X2 = X, Y2 = Y;
					double dX10 = Object.X1-Object.X0, dY10 = Object.Y1-Object.Y0;
					double dX21 = X2-Object.X1, dY21 = Y2-Object.Y1;
					double L = Math.Pow((X2-Object.X1),2)+Math.Pow((Y2-Object.Y1),2);
					//.							
					double Alpha;
					if (dX10 != 0)
					{
						Alpha = Math.Atan(dY10/dX10);
						if (dX10 < 0) Alpha = Alpha+Math.PI;
					}
					else
					{
						if (dY10 >= 0) Alpha = Math.PI/2; else Alpha = -Math.PI/2;
					};
					double Betta;
					if (dX21 != 0)
					{
						Betta = Math.Atan(dY21/dX21);
						if (dX21 < 0) Betta = Betta+Math.PI;
					}
					else
					{
						if (dY21 >= 0) Betta = Math.PI/2; else Betta = -Math.PI/2;
					}
					double Angle = -(Alpha-Betta);
					//.
					ITBase2DVisualizationSOAPFunctionalityService VF = new ITBase2DVisualizationSOAPFunctionalityService(devicesServer.Configuration.SpaceSOAPServerURL);
					if (L > Object.L) 
					{
						VF.Rotate(User.UserName,User.UserPassword,Object.idTVisualization,Object.idVisualization, Object.X1,Object.Y1, Angle);
						VF.SetPosition(User.UserName,User.UserPassword,Object.idTVisualization,Object.idVisualization, X,Y,0);
						//.
						Object.X0 = Object.X1; Object.Y0 = Object.Y1;
						Object.X1 = X2; Object.Y1 = Y2;
					};
				};
			};
			//.
			this.Invoke(new MeasurementReceivedEvent(UpdateOnMeasurementReceived),new object[1] {Message});
		}

		private void btnShowUsers_Click(object sender, System.EventArgs e)
		{
			TfmGEOGraphUsers fm = new TfmGEOGraphUsers(devicesServer.Users);
			try
			{
				fm.ShowDialog();
			}
			finally
			{
				fm.Dispose();
			};
		}

		[STAThread]
		static void Main() 
		{
			Application.Run(new TfmDevicesServer());
		}

		private void cbProcessing_CheckedChanged(object sender, System.EventArgs e)
		{
			if (!cbProcessing.Checked)
			{
				cbProcessing.BackColor = Color.Red;
			}
			else
			{
				cbProcessing.BackColor = this.BackColor;
			};
		}

		private void cbLogEnabled_CheckedChanged(object sender, System.EventArgs e)
		{
			if (cbLogEnabled.Checked)
			{
				lvLog.Enabled = true;
			}
			else
			{
				lvLog.Enabled = false;
			};
		}
	}

	class DevicesServer
	{
		public TServerConfiguration Configuration = new TServerConfiguration();
		public string RootFolder = "";
		public TfmDevicesServer fmDevicesServer = null;
		private Thread Listening;
		private Socket socket = null;
		private Socket client = null;
		public MeasurementReceivedEvent MeasurementReceived = null;
		public TGEOGraphUsers Users;

		public DevicesServer(TfmDevicesServer pfmDevicesServer)
		{
			fmDevicesServer = pfmDevicesServer;
			Configuration.Load();
			MeasurementReceived = new MeasurementReceivedEvent(fmDevicesServer.DoOnMeasurementReceived);
			Users = new TGEOGraphUsers();
			ThreadStart ThreadDelegate = new ThreadStart(Listen);
			Listening = new Thread(ThreadDelegate);
			Listening.Start();
		}

		public void Dispose()
		{
			if (Listening != null) {socket.Close(); Listening.Abort();};
		}

		private void Listen()
		{
			IPEndPoint ipep = new IPEndPoint(IPAddress.Any, Configuration.Port);
			socket = new Socket(AddressFamily.InterNetwork,SocketType.Stream, ProtocolType.Tcp);
			try
			{
				socket.Bind(ipep);
				socket.Listen(Configuration.MaxConnections);
				while (true) 
				{
					client = socket.Accept();
					ClientThread ClientProcessing = new ClientThread(this,client);
				}
			}
			finally 
			{
				socket.Close();
			}
		}

	}

	class ClientThread
	{
		private DevicesServer Server;
		private Socket client;
		private Thread Processing;

		public ClientThread(DevicesServer pServer, Socket pclient)
		{
			Server = pServer;
			client = pclient;
			ThreadStart ThreadDelegate = new ThreadStart(DoProcess);
			Processing = new Thread(ThreadDelegate);
			Processing.Start();
		}

		private void DoProcess() 
		{
			try
			{
				IPEndPoint clientep = (IPEndPoint)client.RemoteEndPoint;
				//. get device preamble
				byte[] BA = new byte[GEOGraphMessage.EncryptedMessageSize];
				int RetSize = client.Receive(BA);
				if (RetSize != GEOGraphMessage.EncryptedMessageSize) throw new Exception("  "); //. =>
				GEOGraphMessage Message = new GEOGraphMessage();
				Message.Preset(BA);
				TGEOGraphUser User;
				TGEOGraphUserObject Object;
				string UserPassword = null;
				lock(Server.Users)
				{
					if (Server.Users.GetUserAndObjectByObjectID(Message.ObjectID, out User, out Object)) UserPassword = User.UserPassword;
				};
				if (UserPassword != null)
				{
					Message.Decript(UserPassword);
					if (Server.MeasurementReceived != null) Server.MeasurementReceived(Message);
				};
			}
			finally 
			{
				client.Close();
			}
		}
	}

	public struct GEOGraphMessage
	{
		public static int EncryptedMessageSize = 38;

		public byte[] BA;
		//.
		public byte Type;
		public byte Version;
		public int ObjectID;
		public double Latitude;
		public double Longitude;
		public double Precision;

		public void Preset(byte[] pBA)
		{
			BA = pBA;
			int Idx = 0;
			Type = BA[Idx]; Idx++;
			Version = BA[Idx]; Idx++;
			ObjectID = BitConverter.ToInt32(BA,Idx); 
		}

		public void Decript(string KeyString)
		{
			MessageDecryptor Decryptor = new MessageDecryptor(KeyString);
			try
			{
				byte[] DBA = Decryptor.Decrypt(BA,1+1+4,BA.Length-6);
				int Idx = 0;
				Latitude = BitConverter.ToDouble(DBA,Idx);		  Idx+=8;
				Longitude = BitConverter.ToDouble(DBA,Idx);		  Idx+=8;
				Precision = BitConverter.ToDouble(DBA,Idx); 
			}
			finally
			{
				Decryptor.Destroy();
			};
		}
	}

	public class MessageDecryptor
	{
		private ICryptoTransform Decryptor;

		public MessageDecryptor(string pKeyString)
		{
			System.Text.Encoding encoding = System.Text.Encoding.ASCII;
			byte[] KeyStringBytes = encoding.GetBytes(pKeyString);
			byte[] MD5Hash;
			using (MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider())
			{
				MD5Hash = md5.ComputeHash(KeyStringBytes);	
			};
			using (RC2CryptoServiceProvider rc2 = new RC2CryptoServiceProvider())
			{
				rc2.IV = new byte[8] {255,1,128,2,64,4,32,8};
				rc2.Key = MD5Hash;
				Decryptor = rc2.CreateDecryptor();
			};
		}

		public void Destroy()
		{
			Decryptor.Dispose();
		}

		public byte[] Decrypt(byte[] Message, int Idx, int Size)
		{
			MemoryStream RS = new MemoryStream(Size);
			try
			{
				RS.Write(Message,Idx,Size); RS.Position = 0;
				CryptoStream DecryptiongStream = new CryptoStream(RS,Decryptor,CryptoStreamMode.Read);
				try
				{
					byte[] BA = new byte[Size];
					int BytesCount = DecryptiongStream.Read(BA,0,Size);
					byte[] Result = new byte[BytesCount];
					Array.Copy(BA,0,Result,0,BytesCount);
					return Result; //. ->
				}
				finally
				{
					DecryptiongStream.Close();
				};
			}
			finally
			{
				RS.Close();
			};
		}
	}

	public struct TServerConfiguration
	{
		public int Port;
		public int MaxConnections;
		public string SpaceSOAPServerURL;

		public void Load()
		{
			IniFile ini = new IniFile(Path.GetDirectoryName(Application.ExecutablePath)+@"\Config.ini");
			//. port
			Port = Convert.ToInt32(ini.IniReadValue("START","Port"));
			MaxConnections = Convert.ToInt32(ini.IniReadValue("START","MaxConnections"));
			SpaceSOAPServerURL = ini.IniReadValue("START","SpaceSOAPServerURL");
		}
	}

	public class IniFile
	{
		public string path;
		[DllImport("kernel32")]
		private static extern long WritePrivateProfileString(string section, string key,string val,string filePath);
		[DllImport("kernel32")]		
		private static extern int GetPrivateProfileString(string section, string key,string def, StringBuilder retVal,int size,string filePath);
		public IniFile(string INIPath)
		{
			path = INIPath;
		}
		public void IniWriteValue(string Section,string Key,string Value)
		{
			WritePrivateProfileString(Section,Key,Value,this.path);
		}
        
		public string IniReadValue(string Section,string Key)
		{
			StringBuilder temp = new StringBuilder(255);
			int i = GetPrivateProfileString(Section,Key,"",temp, 
				255, this.path);
			return temp.ToString();

		}
	}
}
